# we will use prov-o to represent the provenance of the neat graph store
# basically tracking changes that occur in the graph store
# prov-o use concepts of Agent, Activity and Entity to represent provenance
# where in case of neat we have:
# Agent: triples extractors, graph enhancers, contextualizers, etc.
# Activity: write triple, add connection, etc.
# Entity: neat graph store


import uuid
from collections import UserList
from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime
from typing import TypeVar

from rdflib import PROV, RDF, Literal, URIRef

from cognite.neat.constants import DEFAULT_NAMESPACE


@dataclass(frozen=True)
class Agent:
    id_: URIRef = DEFAULT_NAMESPACE.agent
    acted_on_behalf_of: str = "NEAT"

    def as_triples(self):
        return [
            (self.id_, RDF.type, PROV[type(self).__name__]),
            (self.id_, PROV.actedOnBehalfOf, self.acted_on_behalf_of),
        ]


@dataclass(frozen=True)
class Activity:
    was_associated_with: Agent
    ended_at_time: datetime
    started_at_time: datetime
    used: str  # this would be set to for example Extractor, Enhancer, Contextualizer, etc.
    id_: URIRef = DEFAULT_NAMESPACE[f"activity-{uuid.uuid4()}"]

    def as_triples(self):
        return [
            (self.id_, RDF.type, PROV[type(self).__name__]),
            (self.id_, PROV.wasAssociatedWith, self.was_associated_with.id_),
            (self.id_, PROV.startedAtTime, Literal(self.started_at_time)),
            (self.id_, PROV.endedAtTime, Literal(self.ended_at_time)),
            (self.id_, PROV.used, self.used),
        ]


@dataclass(frozen=True)
class Entity:
    was_generated_by: Activity
    was_attributed_to: Agent
    id_: URIRef = DEFAULT_NAMESPACE["graph-store"]

    def as_triples(self):
        return [
            (self.id_, RDF.type, PROV[type(self).__name__]),
            (self.id_, PROV.wasGeneratedBy, self.was_generated_by.id_),
            (self.id_, PROV.wasAttributedTo, self.was_attributed_to.id_),
        ]


@dataclass(frozen=True)
class Change:
    agent: Agent
    activity: Activity
    entity: Entity
    description: str

    def as_triples(self):
        return self.agent.as_triples() + self.activity.as_triples() + self.entity.as_triples()

    @classmethod
    def record(cls, activity: str, start: datetime, end: datetime, description: str):
        """User friendly method to record a change that occurred in the graph store."""
        agent = Agent()
        activity = Activity(used=activity, was_associated_with=agent, started_at_time=start, ended_at_time=end)
        entity = Entity(was_generated_by=activity, was_attributed_to=agent)
        return cls(agent, activity, entity, description)


T_Change = TypeVar("T_Change", bound=Change)


class Provenance(UserList[T_Change]):
    def __init__(self, changes: Sequence[T_Change] | None = None):
        super().__init__(changes or [])

    def activity_took_place(self, activity: str) -> bool:
        return any(change.activity.used == activity for change in self)

    def __delitem__(self, *args, **kwargs):
        raise TypeError("Cannot delete change from provenance")

    def __setitem__(self, *args, **kwargs):
        raise TypeError("Cannot modify change from provenance")



from abc import ABC
from dataclasses import dataclass, field
from functools import total_ordering

from cognite.neat.issues import NeatIssueList


@total_ordering
@dataclass
class UploadResultCore(ABC):
    name: str
    error_messages: list[str] = field(default_factory=list)
    issues: NeatIssueList = field(default_factory=NeatIssueList)

    def __lt__(self, other: object) -> bool:
        if isinstance(other, UploadDiffsCount):
            return self.name < other.name
        else:
            return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, UploadDiffsCount):
            return self.name == other.name
        else:
            return NotImplemented


@dataclass
class UploadDiffsCount(UploadResultCore):
    created: int = 0
    deleted: int = 0
    changed: int = 0
    unchanged: int = 0
    skipped: int = 0
    failed_created: int = 0
    failed_changed: int = 0
    failed_deleted: int = 0

    @property
    def total(self) -> int:
        return self.created + self.deleted + self.changed + self.unchanged

    @property
    def failed(self) -> int:
        return self.failed_created + self.failed_changed + self.failed_deleted

    def as_report_str(self) -> str:
        line = []
        if self.created:
            line.append(f"created {self.created}")
        if self.changed:
            line.append(f"updated {self.changed}")
        if self.skipped:
            line.append(f"skipped {self.skipped}")
        if self.unchanged:
            line.append(f"unchanged {self.unchanged}")
        if self.deleted:
            line.append(f"deleted {self.deleted}")
        if self.failed_created:
            line.append(f"failed to create {self.failed_created}")
        if self.failed_changed:
            line.append(f"failed to update {self.failed_changed}")
        if self.failed_deleted:
            line.append(f"failed to delete {self.failed_deleted}")

        return f"{self.name.title()}: {', '.join(line)}"


@dataclass
class UploadResultIDs(UploadResultCore):
    success: list[str] = field(default_factory=list)
    failed: list[str] = field(default_factory=list)


@dataclass
class UploadDiffsID(UploadResultCore):
    created: list[str] = field(default_factory=list)
    changed: list[str] = field(default_factory=list)
    unchanged: list[str] = field(default_factory=list)
    failed: list[str] = field(default_factory=list)

    def as_upload_result_ids(self) -> UploadResultIDs:
        result = UploadResultIDs(name=self.name, error_messages=self.error_messages, issues=self.issues)
        result.success = self.created + self.changed + self.unchanged
        result.failed = self.failed
        return result



from collections.abc import Iterable
from datetime import datetime, timezone
from pathlib import Path
from typing import cast

from cognite.client import CogniteClient
from cognite.client.data_classes import LabelDefinition, LabelDefinitionList
from rdflib import RDF, Literal, Namespace

from cognite.neat.constants import DEFAULT_NAMESPACE
from cognite.neat.graph.extractors._base import BaseExtractor
from cognite.neat.graph.models import Triple
from cognite.neat.utils.utils import create_sha256_hash


class LabelsExtractor(BaseExtractor):
    def __init__(
        self,
        labels: Iterable[LabelDefinition],
        namespace: Namespace | None = None,
    ):
        self.namespace = namespace or DEFAULT_NAMESPACE
        self.labels = labels

    @classmethod
    def from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
        namespace: Namespace | None = None,
    ):
        return cls(
            cast(Iterable[LabelDefinition], client.labels(data_set_external_ids=data_set_external_id)), namespace
        )

    @classmethod
    def from_file(cls, file_path: str, namespace: Namespace | None = None):
        return cls(LabelDefinitionList.load(Path(file_path).read_text()), namespace)

    def extract(self) -> Iterable[Triple]:
        """Extract labels as triples."""
        for label in self.labels:
            yield from self._labels2triples(label, self.namespace)

    @classmethod
    def _labels2triples(cls, label: LabelDefinition, namespace: Namespace) -> list[Triple]:
        if label.external_id:
            id_ = namespace[f"Label_{create_sha256_hash(label.external_id)}"]

            # Set rdf type
            triples: list[Triple] = [(id_, RDF.type, namespace.Label)]

            # Create attributes
            triples.append((id_, namespace.external_id, Literal(label.external_id)))

            if label.name:
                triples.append((id_, namespace.name, Literal(label.name)))

            if label.description:
                triples.append((id_, namespace.description, Literal(label.description)))

            if label.created_time:
                triples.append(
                    (
                        id_,
                        namespace.created_time,
                        Literal(datetime.fromtimestamp(label.created_time / 1000, timezone.utc)),
                    )
                )

            if label.data_set_id:
                triples.append((id_, namespace.data_set_id, namespace[f"Dataset_{label.data_set_id}"]))

            return triples
        return []



from ._classic_cdf._assets import AssetsExtractor
from ._classic_cdf._events import EventsExtractor
from ._classic_cdf._files import FilesExtractor
from ._classic_cdf._labels import LabelsExtractor
from ._classic_cdf._relationships import RelationshipsExtractor
from ._classic_cdf._sequences import SequencesExtractor
from ._classic_cdf._timeseries import TimeSeriesExtractor
from ._dexpi import DexpiExtractor
from ._mock_graph_generator import MockGraphGenerator
from ._rdf_file import RdfFileExtractor

__all__ = [
    "AssetsExtractor",
    "MockGraphGenerator",
    "RelationshipsExtractor",
    "TimeSeriesExtractor",
    "SequencesExtractor",
    "EventsExtractor",
    "FilesExtractor",
    "LabelsExtractor",
    "RdfFileExtractor",
    "DexpiExtractor",
]


TripleExtractors = (
    AssetsExtractor
    | MockGraphGenerator
    | RelationshipsExtractor
    | TimeSeriesExtractor
    | SequencesExtractor
    | EventsExtractor
    | FilesExtractor
    | LabelsExtractor
    | RdfFileExtractor
    | DexpiExtractor
)



from collections.abc import Iterable
from datetime import datetime, timezone
from pathlib import Path
from typing import cast
from urllib.parse import quote

from cognite.client import CogniteClient
from cognite.client.data_classes import FileMetadata, FileMetadataList
from pydantic import AnyHttpUrl, ValidationError
from rdflib import RDF, Literal, Namespace, URIRef

from cognite.neat.constants import DEFAULT_NAMESPACE
from cognite.neat.graph.extractors._base import BaseExtractor
from cognite.neat.graph.models import Triple
from cognite.neat.utils.utils import string_to_ideal_type


class FilesExtractor(BaseExtractor):
    def __init__(
        self,
        events: Iterable[FileMetadata],
        namespace: Namespace | None = None,
    ):
        self.namespace = namespace or DEFAULT_NAMESPACE
        self.events = events

    @classmethod
    def from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
        namespace: Namespace | None = None,
    ):
        return cls(cast(Iterable[FileMetadata], client.files(data_set_external_ids=data_set_external_id)), namespace)

    @classmethod
    def from_file(cls, file_path: str, namespace: Namespace | None = None):
        return cls(FileMetadataList.load(Path(file_path).read_text()), namespace)

    def extract(self) -> Iterable[Triple]:
        """Extract files metadata as triples."""
        for event in self.events:
            yield from self._file2triples(event, self.namespace)

    @classmethod
    def _file2triples(cls, file: FileMetadata, namespace: Namespace) -> list[Triple]:
        id_ = namespace[f"File_{file.id}"]

        # Set rdf type
        triples: list[Triple] = [(id_, RDF.type, namespace.File)]

        # Create attributes

        if file.external_id:
            triples.append((id_, namespace.external_id, Literal(file.external_id)))

        if file.source:
            triples.append((id_, namespace.type, Literal(file.source)))

        if file.mime_type:
            triples.append((id_, namespace.mime_type, Literal(file.mime_type)))

        if file.uploaded:
            triples.append((id_, namespace.uploaded, Literal(file.uploaded)))

        if file.source:
            triples.append((id_, namespace.source, Literal(file.source)))

        if file.metadata:
            for key, value in file.metadata.items():
                if value:
                    type_aware_value = string_to_ideal_type(value)
                    try:
                        triples.append((id_, namespace[key], URIRef(str(AnyHttpUrl(type_aware_value)))))  # type: ignore
                    except ValidationError:
                        triples.append((id_, namespace[key], Literal(type_aware_value)))

        if file.source_created_time:
            triples.append(
                (
                    id_,
                    namespace.source_created_time,
                    Literal(datetime.fromtimestamp(file.source_created_time / 1000, timezone.utc)),
                )
            )
        if file.source_modified_time:
            triples.append(
                (
                    id_,
                    namespace.source_created_time,
                    Literal(datetime.fromtimestamp(file.source_modified_time / 1000, timezone.utc)),
                )
            )
        if file.uploaded_time:
            triples.append(
                (id_, namespace.uploaded_time, Literal(datetime.fromtimestamp(file.uploaded_time / 1000, timezone.utc)))
            )

        if file.created_time:
            triples.append(
                (id_, namespace.created_time, Literal(datetime.fromtimestamp(file.created_time / 1000, timezone.utc)))
            )

        if file.last_updated_time:
            triples.append(
                (
                    id_,
                    namespace.last_updated_time,
                    Literal(datetime.fromtimestamp(file.last_updated_time / 1000, timezone.utc)),
                )
            )

        if file.labels:
            for label in file.labels:
                # external_id can create ill-formed URIs, so we create websafe URIs
                # since labels do not have internal ids, we use the external_id as the id
                triples.append((id_, namespace.label, namespace[f"Label_{quote(label.dump()['externalId'])}"]))

        if file.security_categories:
            for category in file.security_categories:
                triples.append((id_, namespace.security_categories, Literal(category)))

        if file.data_set_id:
            triples.append((id_, namespace.data_set_id, namespace[f"Dataset_{file.data_set_id}"]))

        if file.asset_ids:
            for asset_id in file.asset_ids:
                triples.append((id_, namespace.asset, namespace[f"Asset_{asset_id}"]))

        return triples



from pathlib import Path

from rdflib import URIRef

from cognite.neat.graph._shared import MIMETypes
from cognite.neat.graph.extractors._base import BaseExtractor


class RdfFileExtractor(BaseExtractor):
    def __init__(
        self,
        filepath: Path,
        mime_type: MIMETypes = "application/rdf+xml",
        base_uri: URIRef | None = None,
    ):
        self.filepath = filepath
        self.mime_type = mime_type
        self.base_uri = base_uri



from abc import ABC, abstractmethod
from collections.abc import Iterable
from pathlib import Path
from typing import ClassVar, Generic, Literal, TypeVar, overload

from cognite.client import CogniteClient
from cognite.client.data_classes.capabilities import Capability

from cognite.neat.graph import NeatGraphStore
from cognite.neat.graph.issues.loader import FailedAuthorizationError
from cognite.neat.issues import NeatIssue, NeatIssueList
from cognite.neat.utils.upload import UploadDiffsID, UploadResultIDs

T_Output = TypeVar("T_Output")


class BaseLoader(ABC, Generic[T_Output]):
    _new_line = "\n"
    _encoding = "utf-8"

    def __init__(self, graph_store: NeatGraphStore):
        self.graph_store = graph_store

    @abstractmethod
    def write_to_file(self, filepath: Path) -> None:
        raise NotImplementedError

    def load(self, stop_on_exception: bool = False) -> Iterable[T_Output | NeatIssue]:
        """Load the graph with data."""
        return self._load(stop_on_exception)

    @abstractmethod
    def _load(self, stop_on_exception: bool = False) -> Iterable[T_Output | NeatIssue]:
        """Load the graph with data."""
        pass


class CDFLoader(BaseLoader[T_Output]):
    _UPLOAD_BATCH_SIZE: ClassVar[int] = 1000

    @overload
    def load_into_cdf_iterable(
        self, client: CogniteClient, return_diffs: Literal[False] = False, dry_run: bool = False
    ) -> Iterable[UploadResultIDs]: ...

    @overload
    def load_into_cdf_iterable(
        self, client: CogniteClient, return_diffs: Literal[True], dry_run: bool = False
    ) -> Iterable[UploadDiffsID]: ...

    def load_into_cdf_iterable(
        self, client: CogniteClient, return_diffs: bool = False, dry_run: bool = False
    ) -> Iterable[UploadResultIDs] | Iterable[UploadDiffsID]:
        yield from self._load_into_cdf_iterable(client, return_diffs, dry_run)

    @overload
    def load_into_cdf(
        self, client: CogniteClient, return_diffs: Literal[False] = False, dry_run: bool = False
    ) -> list[UploadResultIDs]: ...

    @overload
    def load_into_cdf(
        self, client: CogniteClient, return_diffs: Literal[True], dry_run: bool = False
    ) -> list[UploadDiffsID]: ...

    def load_into_cdf(
        self, client: CogniteClient, return_diffs: bool = False, dry_run: bool = False
    ) -> list[UploadResultIDs] | list[UploadDiffsID]:
        return list(self._load_into_cdf_iterable(client, return_diffs, dry_run))  # type: ignore[return-value]

    def _load_into_cdf_iterable(
        self, client: CogniteClient, return_diffs: bool = False, dry_run: bool = False
    ) -> Iterable[UploadResultIDs] | Iterable[UploadDiffsID]:
        missing_capabilities = client.iam.verify_capabilities(self._get_required_capabilities())
        result_cls = UploadDiffsID if return_diffs else UploadResultIDs
        if missing_capabilities:
            result = result_cls(name=type(self).__name__)
            result.issues.append(FailedAuthorizationError(action="Upload to CDF", reason=str(missing_capabilities)))
            yield result
            return

        issues = NeatIssueList[NeatIssue]()
        items: list[T_Output] = []
        for result in self.load(stop_on_exception=False):
            if isinstance(result, NeatIssue):
                issues.append(result)
            else:
                items.append(result)

            if len(items) >= self._UPLOAD_BATCH_SIZE:
                yield self._upload_to_cdf(client, items, return_diffs, dry_run, issues)
                items.clear()
        if items:
            yield self._upload_to_cdf(client, items, return_diffs, dry_run, issues)

    @abstractmethod
    def _get_required_capabilities(self) -> list[Capability]:
        raise NotImplementedError

    @abstractmethod
    def _upload_to_cdf(
        self,
        client: CogniteClient,
        items: list[T_Output],
        return_diffs: bool,
        dry_run: bool,
        read_issues: NeatIssueList,
    ) -> UploadResultIDs | UploadDiffsID:
        raise NotImplementedError
